------------------------------------------------------------------------
-- Equality coercions (structural equalities are also β/η equalities)
------------------------------------------------------------------------

-- One can of course turn an equality which doesn't use the β/η axioms
-- into one which (potentially) does.

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import Implicit.Equality
-- # import EqKind
-- # import TmVar

abstract

  coerceToTm⁻ (k : EqKind)
              (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
              (t⁻¹ : Tm⁻ tv¹ t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
              (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq βη t⁻¹ t⁻²
  coerceToTm⁻ (var⁻Cong eqv)       = var⁻Cong eqv
  coerceToTm⁻ (λ⁻Cong eqt)         = λ⁻Cong $ coerceToTm⁻ eqt
  coerceToTm⁻ (app⁻Cong eqt₁ eqt₂) =
    app⁻Cong (coerceToTm⁻ eqt₁) (coerceToTm⁻ eqt₂)
  coerceToTm⁻ (castTm⁻Eqˡ eqt eqt⁻) =
    castTm⁻Eqˡ eqt (coerceToTm⁻ eqt⁻)
  coerceToTm⁻ (castTm⁻Eqʳ eqt eqt⁻) =
    castTm⁻Eqʳ eqt (coerceToTm⁻ eqt⁻)
  coerceToTm⁻ (symTm⁻βη eqt⁻)            = symTm⁻βη eqt⁻
  coerceToTm⁻ (transTm⁻βη eqt⁻¹² eqt⁻²³) = transTm⁻βη eqt⁻¹² eqt⁻²³
  coerceToTm⁻ (β⁻ t⁻₁ t⁻₂)               = β⁻ t⁻₁ t⁻₂
  coerceToTm⁻ (η⁻ t⁻)                    = η⁻ t⁻

  coerceFromTm⁻ (k : EqKind)
                (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
                (t⁻¹ : Tm⁻ tv¹ t¹)
                (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
                (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq structural t⁻¹ t⁻² -> Tm⁻Eq k t⁻¹ t⁻²
  coerceFromTm⁻ {structural} eqt = eqt
  coerceFromTm⁻ {βη}         eqt = coerceToTm⁻ eqt

  coerceToSubst⁻ (k : EqKind)
                 (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                 (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                 (tv² : TmVar) (Γ², Δ² : Ctxt)
                 (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : Subst⁻Eq k ρ⁻¹ ρ⁻² -> Subst⁻Eq βη ρ⁻¹ ρ⁻²
  coerceToSubst⁻ (extEq⁻ eqΓ eqΔ proof) =
    extEq⁻ eqΓ eqΔ
           (\_ _ _ _ eqv -> coerceToTm⁻ (proof _ _ _ _ eqv))

  coerceFromSubst⁻ (k : EqKind)
                   (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                   (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                   (tv² : TmVar) (Γ², Δ² : Ctxt)
                   (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : Subst⁻Eq structural ρ⁻¹ ρ⁻² -> Subst⁻Eq k ρ⁻¹ ρ⁻²
  coerceFromSubst⁻ {structural} eqρ = eqρ
  coerceFromSubst⁻ {βη}         eqρ = coerceToSubst⁻ eqρ

Implicit-Coercions : Type
Implicit-Coercions = time-plugin(Implicit-Coercions, Set)
