------------------------------------------------------------------------
-- Miscellaneous equality lemmas
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import Implicit.Equality
-- # import Implicit.Equivalence
-- # import EqKind
-- # import TmVar

------------------------------------------------------------------------
-- Some equality lemmas and congruences

abstract

  castTm⁻Eq (k : EqKind) (tv : TmVar) (Γ : Ctxt)
            (τ¹ : Ty Γ) (t¹ : Γ ⊢ τ¹)
            (τ² : Ty Γ) (t² : Γ ⊢ τ²)
    : (t⁻ : Tm⁻ tv t¹) -> (eqt : t¹ ≈⊢ t²) -> Tm⁻Eq k (t⁻ ∷≡⊢⁻ eqt) t⁻
  castTm⁻Eq t⁻ eqt = castTm⁻Eqˡ eqt (reflTm⁻ t⁻)

  betweenTm⁻ (k : EqKind)
             (Γ₁ : Ctxt) (τ₁ : Ty Γ₁) (t₁ : Γ₁ ⊢ τ₁) (tv₁ : TmVar)
             (t⁻₁ : Tm⁻ tv₁ t₁)
             (Γ₂ : Ctxt) (τ₂ : Ty Γ₂) (t₂ : Γ₂ ⊢ τ₂) (tv₂ : TmVar)
             (t⁻₂ : Tm⁻ tv₂ t₂)
             (Γ₃ : Ctxt) (τ₃ : Ty Γ₃) (t₃ : Γ₃ ⊢ τ₃) (tv₃ : TmVar)
             (t⁻₃ : Tm⁻ tv₃ t₃)
             (Γ₄ : Ctxt) (τ₄ : Ty Γ₄) (t₄ : Γ₄ ⊢ τ₄) (tv₄ : TmVar)
             (t⁻₄ : Tm⁻ tv₄ t₄)
    :  Tm⁻Eq k t⁻₁ t⁻₂ -> Tm⁻Eq k t⁻₄ t⁻₃ -> Tm⁻Eq k t⁻₂ t⁻₃
    -> Tm⁻Eq k t⁻₁ t⁻₄
  betweenTm⁻ eq₁ eq₃ eq₂ = eq₁ `transTm⁻` eq₂ `transTm⁻` symTm⁻ eq₃


  betweenTy⁻ (Γ₁ : Ctxt) (τ₁ : Ty Γ₁) (τ⁻₁ : Ty⁻ τ₁)
             (Γ₂ : Ctxt) (τ₂ : Ty Γ₂) (τ⁻₂ : Ty⁻ τ₂)
             (Γ₃ : Ctxt) (τ₃ : Ty Γ₃) (τ⁻₃ : Ty⁻ τ₃)
             (Γ₄ : Ctxt) (τ₄ : Ty Γ₄) (τ⁻₄ : Ty⁻ τ₄)
    :  τ⁻₁ ≛⁻ τ⁻₂ -> τ⁻₄ ≛⁻ τ⁻₃ -> τ⁻₂ ≛⁻ τ⁻₃ -> τ⁻₁ ≛⁻ τ⁻₄
  betweenTy⁻ eq₁ eq₃ eq₂ = eq₁ `transTy⁻` eq₂ `transTy⁻` symTy⁻ eq₃

  substVar⁻Cong (k : EqKind)
    (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt) (σ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ σ¹)
    (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
    (tv² : TmVar) (Γ², Δ² : Ctxt) (σ² : Ty Γ²) (v² : Γ² ∋ σ²)
    (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    :  v¹ ≈∋ v² -> Subst⁻Eq k ρ⁻¹ ρ⁻²
    -> Tm⁻Eq k (v¹ /∋⁻ ρ⁻¹) (v² /∋⁻ ρ⁻²)
  substVar⁻Cong eqv (extEq⁻ eqΓ eqΔ proof) = proof _ _ _ _ eqv

  substCastVar⁻ (k : EqKind) (tv : TmVar)
                (Γ, Δ : Ctxt) (τ¹, τ² : Ty Γ) (ρ : Subst Γ Δ)
    :  (v : Γ ∋ τ¹) -> (eqτ : τ¹ ≛ τ²) -> (ρ⁻ : Subst⁻ tv ρ)
    -> Tm⁻Eq k ((v ∷≡∋ eqτ) /∋⁻ ρ⁻) (v /∋⁻ ρ⁻)
  substCastVar⁻ v eqτ ρ⁻ =
    substVar⁻Cong (castVarEq v eqτ) (reflSubst⁻ ρ⁻)

  subst⁻EqCtxtEqLeft (k : EqKind)
                     (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                     (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                     (tv² : TmVar) (Γ², Δ² : Ctxt)
                     (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : Subst⁻Eq k ρ⁻¹ ρ⁻² -> Γ¹ ≜ Γ²
  subst⁻EqCtxtEqLeft (extEq⁻ eqΓ eqΔ proof) = eqΓ

  subst⁻EqCtxtEqRight (k : EqKind)
                      (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt)
                      (ρ¹ : Subst Γ¹ Δ¹) (ρ⁻¹ : Subst⁻ tv¹ ρ¹)
                      (tv² : TmVar) (Γ², Δ² : Ctxt)
                      (ρ² : Subst Γ² Δ²) (ρ⁻² : Subst⁻ tv² ρ²)
    : Subst⁻Eq k ρ⁻¹ ρ⁻² -> Δ¹ ≜ Δ²
  subst⁻EqCtxtEqRight (extEq⁻ eqΓ eqΔ proof) = eqΔ

  castSubst⁻Eq (k : EqKind) (tv : TmVar)
               (Γ, Δ : Ctxt) (ρ¹, ρ² : Subst Γ Δ)
    :  (ρ⁻ : Subst⁻ tv ρ¹) -> (eqρ : ρ¹ ≘ ρ²)
    -> Subst⁻Eq k (ρ⁻ ∷≡⇒⁻ eqρ) ρ⁻
  castSubst⁻Eq ρ⁻ eqρ = castSubst⁻Eqˡ eqρ (reflSubst⁻ ρ⁻)

  betweenSubst⁻ (k : EqKind)
                (tv₁ : TmVar) (Γ₁, Δ₁ : Ctxt)
                (ρ₁ : Subst Γ₁ Δ₁) (ρ⁻₁ : Subst⁻ tv₁ ρ₁)
                (tv₂ : TmVar) (Γ₂, Δ₂ : Ctxt)
                (ρ₂ : Subst Γ₂ Δ₂) (ρ⁻₂ : Subst⁻ tv₂ ρ₂)
                (tv₃ : TmVar) (Γ₃, Δ₃ : Ctxt)
                (ρ₃ : Subst Γ₃ Δ₃) (ρ⁻₃ : Subst⁻ tv₃ ρ₃)
                (tv₄ : TmVar) (Γ₄, Δ₄ : Ctxt)
                (ρ₄ : Subst Γ₄ Δ₄) (ρ⁻₄ : Subst⁻ tv₄ ρ₄)
    :  Subst⁻Eq k ρ⁻₁ ρ⁻₂ -> Subst⁻Eq k ρ⁻₄ ρ⁻₃ -> Subst⁻Eq k ρ⁻₂ ρ⁻₃
    -> Subst⁻Eq k ρ⁻₁ ρ⁻₄
  betweenSubst⁻ eq₁ eq₃ eq₂ =
    eq₁ `transSubst⁻` eq₂ `transSubst⁻` symSubst⁻ eq₃

------------------------------------------------------------------------
-- Substitution equality proof variants

-- A simpler variant which is useful when the starting contexts are
-- definitionally equal.

ExtEq⁻ˡ⁼ (Γ, Δ¹, Δ² : Ctxt)
         (tv¹ : TmVar) (ρ¹ : Subst Γ Δ¹)
         (tv² : TmVar) (ρ² : Subst Γ Δ²)
  : EqKind -> Subst⁻ tv¹ ρ¹ -> Subst⁻ tv² ρ² -> Set
ExtEq⁻ˡ⁼ {Γ} k ρ⁻¹ ρ⁻² =
  (σ : Ty Γ) -> (v : Γ ∋ σ) -> Tm⁻Eq k (v /∋⁻ ρ⁻¹) (v /∋⁻ ρ⁻²)

abstract

  -- The substitutions are non-hidden arguments since they often need
  -- to be given to get the code to type check. This would not be the
  -- case if the substitutions were abstract, so that they didn't
  -- evaluate.

  extEq⁻ˡ⁼ (k : EqKind) (Γ : Ctxt)
           (tv¹ : TmVar) (Δ¹ : Ctxt) (ρ¹ : Subst Γ Δ¹)
           (tv² : TmVar) (Δ² : Ctxt) (ρ² : Subst Γ Δ²)
    :  (ρ⁻¹ : Subst⁻ tv¹ ρ¹) -> (ρ⁻² : Subst⁻ tv² ρ²)
    -> Δ¹ ≜ Δ² -> ExtEq⁻ˡ⁼ k ρ⁻¹ ρ⁻² -> Subst⁻Eq k ρ⁻¹ ρ⁻²
  extEq⁻ˡ⁼ {k} {Γ} {tv¹} {Δ¹} {ρ¹} {tv²} {Δ²} {ρ²}
           ρ⁻¹ ρ⁻² eqΔ proof =
    extEq⁻ (reflCtxt Γ) eqΔ
           (\σ¹ v¹ σ² v² eqv ->
               proof σ¹ v¹
                 `transTm⁻`
               substVar⁻Cong eqv (reflSubst⁻ ρ⁻²))

  -- Another variant, useful when also the resulting contexts are
  -- definitionally equal.

  extEq⁻⁼ (k : EqKind) (Γ, Δ : Ctxt)
          (tv¹ : TmVar) (ρ¹ : Subst Γ Δ)
          (tv² : TmVar) (ρ² : Subst Γ Δ)
    :  (ρ⁻¹ : Subst⁻ tv¹ ρ¹) -> (ρ⁻² : Subst⁻ tv² ρ²)
    -> ExtEq⁻ˡ⁼ k ρ⁻¹ ρ⁻² -> Subst⁻Eq k ρ⁻¹ ρ⁻²
  extEq⁻⁼ {k} {Γ} {Δ} ρ⁻¹ ρ⁻² proof =
    extEq⁻ˡ⁼ ρ⁻¹ ρ⁻² (reflCtxt Δ) proof

------------------------------------------------------------------------
-- toTerm/fromVar/Tm⁻⊔ preserve the semantics of their arguments

abstract

  toTermEq (k : EqKind) (tv : TmVar) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (t⁻ : Tm⁻ tv t) -> Tm⁻Eq k (toTerm t⁻) t⁻
  toTermEq (var⁻ tv v)   = var⁻Cong $ reflVar v
  toTermEq (t⁻ ∷≡⊢⁻ eqt) = castTm⁻Cong eqt eqt $ toTermEq t⁻
  toTermEq (λ⁻ _ _)      = reflTm⁻ _
  toTermEq (_ @⁻ _)      = reflTm⁻ _

  toTermCong (k : EqKind)
             (tv¹ : TmVar) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
             (t⁻¹ : Tm⁻ tv¹ t¹)
             (tv² : TmVar) (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
             (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq k (toTerm t⁻¹) (toTerm t⁻²)
  toTermCong eqt⁻ = betweenTm⁻ (toTermEq _) (toTermEq _) eqt⁻

  fromVarEq (k : EqKind) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (tv : TmVar) -> (t⁻ : Tm⁻ Var t) -> Tm⁻Eq k (fromVar tv t⁻) t⁻
  fromVarEq Var              t⁻ = reflTm⁻ t⁻
  fromVarEq (Term' Var refl) t⁻ = toTermEq t⁻

  fromVarCong (k : EqKind)
              (tv¹ : TmVar) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
              (t⁻¹ : Tm⁻ Var t¹)
              (tv² : TmVar) (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
              (t⁻² : Tm⁻ Var t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq k (fromVar tv¹ t⁻¹) (fromVar tv² t⁻²)
  fromVarCong eqt⁻ = betweenTm⁻ (fromVarEq _ _) (fromVarEq _ _) eqt⁻

  Tm⁻⊔Eq (k : EqKind) (tv₂ : TmVar) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (tv₁ : TmVar) -> (t⁻ : Tm⁻ tv₂ t) -> Tm⁻Eq k (Tm⁻⊔ tv₁ t⁻) t⁻
  Tm⁻⊔Eq Var              t⁻ = reflTm⁻ t⁻
  Tm⁻⊔Eq (Term' Var refl) t⁻ = toTermEq t⁻

  Tm⁻⊔Cong (k : EqKind)
           (tv₂¹ : TmVar) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
           (t⁻¹ : Tm⁻ tv₂¹ t¹)
           (tv₂² : TmVar) (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
           (t⁻² : Tm⁻ tv₂² t²)
    :  (tv₁¹, tv₁² : TmVar)
    -> Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq k (Tm⁻⊔ tv₁¹ t⁻¹) (Tm⁻⊔ tv₁² t⁻²)
  Tm⁻⊔Cong _ _ eqt⁻ = betweenTm⁻ (Tm⁻⊔Eq _ _) (Tm⁻⊔Eq _ _) eqt⁻

Implicit-EqualityLemmas : Type
Implicit-EqualityLemmas = time-plugin(Implicit-EqualityLemmas, Set)
