------------------------------------------------------------------------
-- Alternative, more traditional equality for Implicit
------------------------------------------------------------------------

-- I don't want to write (too much) about the Tm⁻ equality in Equality
-- in my paper, so instead I prove here that the equality is
-- equivalent to a more traditional one.

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import Implicit.Equality
-- # import Implicit.Equivalence
-- # import Implicit.EqualityLemmas
-- # import EqKind
-- # import TmVar

infix 40 ≈⊢⁻′

------------------------------------------------------------------------
-- "Traditional" equality

mutual

  (≈⊢⁻′) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
        (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
    : Tm⁻ tv¹ t¹ -> Tm⁻ tv² t² -> Set
  (≈⊢⁻′) = Tm⁻Eq'' _ _ _ _ _ _ _ _

  data Tm⁻Eq''
   :  (Γ¹ : Ctxt) -> (τ¹ : Ty Γ¹) -> (t¹ : Γ¹ ⊢ τ¹) -> (tv¹ : TmVar)
   -> (Γ² : Ctxt) -> (τ² : Ty Γ²) -> (t² : Γ² ⊢ τ²) -> (tv² : TmVar)
   -> Tm⁻ tv¹ t¹ -> Tm⁻ tv² t² -> Set where

    -- Equivalence relation.

    reflTm⁻' (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (tv : TmVar)
      : (t⁻ : Tm⁻ tv t) -> t⁻ ≈⊢⁻′ t⁻
    symTm⁻' (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
            (t⁻¹ : Tm⁻ tv¹ t¹)
            (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
            (t⁻² : Tm⁻ tv² t²)
      : t⁻¹ ≈⊢⁻′ t⁻² -> t⁻² ≈⊢⁻′ t⁻¹
    transTm⁻' (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
              (t⁻¹ : Tm⁻ tv¹ t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
              (t⁻² : Tm⁻ tv² t²)
              (Γ³ : Ctxt) (τ³ : Ty Γ³) (t³ : Γ³ ⊢ τ³) (tv³ : TmVar)
              (t⁻³ : Tm⁻ tv³ t³)
      : t⁻¹ ≈⊢⁻′ t⁻² -> t⁻² ≈⊢⁻′ t⁻³ -> t⁻¹ ≈⊢⁻′ t⁻³

    -- Congruence.

    var⁻Cong' (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹) (tv¹ : TmVar)
              (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²) (tv² : TmVar)
      : v¹ ≈∋ v² -> var⁻ tv¹ v¹ ≈⊢⁻′ var⁻ tv² v²
    λ⁻Cong'
      (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹)) (tv¹ : TmVar)
      (t¹ : Γ¹ ▻ τ₁¹ ⊢ τ₂¹) (τ⁻₁¹ : Ty⁻ τ₁¹) (t⁻¹ : Tm⁻ tv¹ t¹)
      (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²)) (tv² : TmVar)
      (t² : Γ² ▻ τ₁² ⊢ τ₂²) (τ⁻₁² : Ty⁻ τ₁²) (t⁻² : Tm⁻ tv² t²)
      : t⁻¹ ≈⊢⁻′ t⁻² -> λ⁻ τ⁻₁¹ t⁻¹ ≈⊢⁻′ λ⁻ τ⁻₁² t⁻²
    app⁻Cong' (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
              (t₁¹ : Γ¹ ⊢ Π τ₁¹ τ₂¹) (t₂¹ : Γ¹ ⊢ τ₁¹)
              (tv₁¹, tv₂¹ : TmVar)
              (t⁻₁¹ : Tm⁻ tv₁¹ t₁¹) (t⁻₂¹ : Tm⁻ tv₂¹ t₂¹)
              (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
              (t₁² : Γ² ⊢ Π τ₁² τ₂²) (t₂² : Γ² ⊢ τ₁²)
              (tv₁², tv₂² : TmVar)
              (t⁻₁² : Tm⁻ tv₁² t₁²) (t⁻₂² : Tm⁻ tv₂² t₂²)
      :  t⁻₁¹ ≈⊢⁻′ t⁻₁² -> t⁻₂¹ ≈⊢⁻′ t⁻₂²
      -> t⁻₁¹ @⁻ t⁻₂¹ ≈⊢⁻′ t⁻₁² @⁻ t⁻₂²

    -- Casts.

    castTm⁻Eq' (Γ : Ctxt) (tv : TmVar)
               (τ¹ : Ty Γ) (t¹ : Γ ⊢ τ¹)
               (τ² : Ty Γ) (t² : Γ ⊢ τ²)
      : (t⁻ : Tm⁻ tv t¹) -> (eqt : t¹ ≈⊢ t²) -> t⁻ ∷≡⊢⁻ eqt ≈⊢⁻′ t⁻

    -- Evaluation.

    β⁻' (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (τ⁻₁ : Ty⁻ τ₁)
       (t₁ : Γ ▻ τ₁ ⊢ τ₂) (t₂ : Γ ⊢ τ₁) (tv₁, tv₂ : TmVar)
      :  (t⁻₁ : Tm⁻ tv₁ t₁) -> (t⁻₂ : Tm⁻ tv₂ t₂)
      -> λ⁻ τ⁻₁ t⁻₁ @⁻ t⁻₂ ≈⊢⁻′ t⁻₁ /⊢⁻ sub⁻ t⁻₂

    -- η equality.

    η⁻' (tv₁, tv₂, tv₃ : TmVar) (Γ : Ctxt)
        (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (t : Γ ⊢ Π τ₁ τ₂) (τ⁻₁ : Ty⁻ τ₁)
      :  (t⁻ : Tm⁻ tv₁ t)
      -> λ⁻ τ⁻₁ ((t⁻ /⊢⁻ wk⁻ tv₂ τ₁) @⁻ var⁻ tv₃ (vz τ₁)) ≈⊢⁻′ t⁻

------------------------------------------------------------------------
-- The equalities are equivalent

abstract

  tm⁻EqTm⁻Eq' (k : EqKind)
              (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
              (t⁻¹ : Tm⁻ tv¹ t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
              (t⁻² : Tm⁻ tv² t²)
    : Tm⁻Eq k t⁻¹ t⁻² -> t⁻¹ ≈⊢⁻′ t⁻²
  tm⁻EqTm⁻Eq' (var⁻Cong eqv)       = var⁻Cong' eqv
  tm⁻EqTm⁻Eq' (λ⁻Cong eqt)         = λ⁻Cong' (tm⁻EqTm⁻Eq' eqt)
  tm⁻EqTm⁻Eq' (app⁻Cong eqt₁ eqt₂) =
    app⁻Cong' (tm⁻EqTm⁻Eq' eqt₁) (tm⁻EqTm⁻Eq' eqt₂)
  tm⁻EqTm⁻Eq' (castTm⁻Eqˡ eqt eqt⁻) =
    castTm⁻Eq' _ eqt `transTm⁻'` tm⁻EqTm⁻Eq' eqt⁻
  tm⁻EqTm⁻Eq' (castTm⁻Eqʳ eqt eqt⁻) =
    tm⁻EqTm⁻Eq' eqt⁻ `transTm⁻'` symTm⁻' (castTm⁻Eq' _ eqt)
  tm⁻EqTm⁻Eq' (symTm⁻βη eqt⁻)            = symTm⁻' $ tm⁻EqTm⁻Eq' eqt⁻
  tm⁻EqTm⁻Eq' (transTm⁻βη eqt⁻¹² eqt⁻²³) =
    transTm⁻' (tm⁻EqTm⁻Eq' eqt⁻¹²) (tm⁻EqTm⁻Eq' eqt⁻²³)
  tm⁻EqTm⁻Eq' (β⁻ t⁻₁ t⁻₂) = β⁻' t⁻₁ t⁻₂
  tm⁻EqTm⁻Eq' (η⁻ t⁻)      = η⁻' t⁻

  tm⁻Eq'Tm⁻Eq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
              (t⁻¹ : Tm⁻ tv¹ t¹)
              (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
              (t⁻² : Tm⁻ tv² t²)
    : t⁻¹ ≈⊢⁻′ t⁻² -> t⁻¹ ≈⊢⁻ t⁻²
  tm⁻Eq'Tm⁻Eq (reflTm⁻' t)            = reflTm⁻ t
  tm⁻Eq'Tm⁻Eq (symTm⁻' eqt)           = symTm⁻ (tm⁻Eq'Tm⁻Eq eqt)
  tm⁻Eq'Tm⁻Eq (transTm⁻' eqt¹² eqt²³) = transTm⁻ (tm⁻Eq'Tm⁻Eq eqt¹²)
                                                 (tm⁻Eq'Tm⁻Eq eqt²³)
  tm⁻Eq'Tm⁻Eq (var⁻Cong' eqv)         = var⁻Cong eqv
  tm⁻Eq'Tm⁻Eq (λ⁻Cong' eqt)           = λ⁻Cong (tm⁻Eq'Tm⁻Eq eqt)
  tm⁻Eq'Tm⁻Eq (app⁻Cong' eqt₁ eqt₂)   = app⁻Cong (tm⁻Eq'Tm⁻Eq eqt₁)
                                                 (tm⁻Eq'Tm⁻Eq eqt₂)
  tm⁻Eq'Tm⁻Eq (castTm⁻Eq' t eqt)      = castTm⁻Eq t eqt
  tm⁻Eq'Tm⁻Eq (β⁻' t₁ t₂)             = β⁻ t₁ t₂
  tm⁻Eq'Tm⁻Eq (η⁻' t)                 = η⁻ t

Implicit-TraditionalEquality : Type
Implicit-TraditionalEquality =
  time-plugin(Implicit-TraditionalEquality, Set)
