------------------------------------------------------------------------
-- Type checker
------------------------------------------------------------------------

-- The type checker is able to infer the type of annotated terms
-- without explicit substitutions.

-- # import Explicit
-- # import Implicit
-- # import DecidableEquality
-- # import Eq
-- # import TmVar

-- ʳʷ stands for RaW.

infixl 90 @ʳʷ
infix  40 ⊢Var↓?, ⊢Ty↑?, ⊢Tm↓?, ⊢Tm↑?,
          =Var↓,  =Ty↑,  =Tm↓,  =Tm↑

------------------------------------------------------------------------
-- Raw terms

-- Raw variables (de Bruijn indices).

data Varʳʷ : Set where
  vzʳʷ : Varʳʷ
  vsʳʷ : Varʳʷ -> Varʳʷ

mutual

  -- Raw types.

  data Tyʳʷ : Set where
    ✶ʳʷ  : Tyʳʷ
    Πʳʷ  : Tyʳʷ -> Tyʳʷ -> Tyʳʷ
    Elʳʷ : Tmʳʷ -> Tyʳʷ

  -- Raw terms.

  data Tmʳʷ : Set where
    varʳʷ  : Varʳʷ -> Tmʳʷ
    λʳʷ    : Tyʳʷ -> Tmʳʷ -> Tmʳʷ
    (@ʳʷ)  : Tmʳʷ -> Tmʳʷ -> Tmʳʷ

------------------------------------------------------------------------
-- Erasure

eraseVar (Γ : Ctxt) (τ : Ty Γ) : Γ ∋ τ -> Varʳʷ
eraseVar (vz _)    = vzʳʷ
eraseVar (vs v _)  = vsʳʷ (eraseVar v)
eraseVar (v ∷≡∋ _) = eraseVar v

mutual

  eraseTy⁻ (Γ : Ctxt) (τ : Ty Γ) : Ty⁻ τ -> Tyʳʷ
  eraseTy⁻ (✶⁻ _)        = ✶ʳʷ
  eraseTy⁻ (Π⁻ τ⁻₁ τ⁻₂)  = Πʳʷ (eraseTy⁻ τ⁻₁) (eraseTy⁻ τ⁻₂)
  eraseTy⁻ (El⁻ t⁻)      = Elʳʷ (eraseTm⁻ t⁻)
  eraseTy⁻ (τ⁻ ∷≡✶⁻ eqτ) = eraseTy⁻ τ⁻

  eraseTm⁻ (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (tv : TmVar)
    : Tm⁻ tv t -> Tmʳʷ
  eraseTm⁻ (var⁻ _ v)    = varʳʷ (eraseVar v)
  eraseTm⁻ (λ⁻ τ⁻₁ t⁻)   = λʳʷ (eraseTy⁻ τ⁻₁) (eraseTm⁻ t⁻)
  eraseTm⁻ (t⁻₁ @⁻ t⁻₂)  = eraseTm⁻ t⁻₁ @ʳʷ eraseTm⁻ t⁻₂
  eraseTm⁻ (t⁻ ∷≡⊢⁻ eqt) = eraseTm⁻ t⁻

------------------------------------------------------------------------
-- Type checking

data (⊢Var↓?) (Γ : Ctxt) : Varʳʷ -> Set where
  isVar (τ : Ty Γ) : (v : Γ ∋ τ) -> Γ ⊢Var↓? eraseVar v
  noVar (v : Varʳʷ) : Γ ⊢Var↓? v

data (⊢Ty↑?) (Γ : Ctxt) : Tyʳʷ -> Set where
  isTy : (τ : Ty Γ) -> (τ⁻ : Ty⁻ τ) -> Γ ⊢Ty↑? eraseTy⁻ τ⁻
  noTy (e : Tyʳʷ) : Γ ⊢Ty↑? e

data (⊢Tm↑?) {Γ : Ctxt} (τ : Ty Γ) : Tmʳʷ -> Set where
  isTm (tv : TmVar)
    : (t : Γ ⊢ τ) -> (t⁻ : Tm⁻ tv t) -> τ ⊢Tm↑? eraseTm⁻ t⁻
  noTm (e : Tmʳʷ) : τ ⊢Tm↑? e

data (⊢Tm↓?) (Γ : Ctxt) : Tmʳʷ -> Set where
  isTyTm (tv : TmVar)
    :  (τ : Ty Γ) -> (t : Γ ⊢ τ) -> (t⁻ : Tm⁻ tv t)
    -> Γ ⊢Tm↓? eraseTm⁻ t⁻
  noTyTm (e : Tmʳʷ) : Γ ⊢Tm↓? e

inferVar : (Γ : Ctxt) -> (v : Varʳʷ) -> Γ ⊢Var↓? v
inferVar ε       _        = noVar
inferVar (Γ ▻ τ) vzʳʷ     = isVar (vz τ)
inferVar (Γ ▻ τ) (vsʳʷ v) with inferVar Γ v =>
  inferVar₁ _       _                   noVar     = noVar
  inferVar₁ (Γ ▻ τ) (vsʳʷ (eraseVar v)) (isVar v) = isVar (vs v τ)

mutual

  checkTy : (Γ : Ctxt) -> (e : Tyʳʷ) -> Γ ⊢Ty↑? e
  checkTy Γ ✶ʳʷ         = isTy _ (✶⁻ Γ)
  checkTy Γ (Πʳʷ e₁ e₂) with checkTy Γ e₁ =>
    checkTy₁ _ _                       noTy          = noTy
    checkTy₁ Γ (Πʳʷ (eraseTy⁻ τ⁻₁) e₂) (isTy τ₁ τ⁻₁)
      with checkTy (Γ ▻ τ₁) e₂ =>
        checkTy₂ _ _ _ noTy = noTy
        checkTy₂ Γ (Πʳʷ (eraseTy⁻ τ⁻₁) (eraseTy⁻ τ⁻₂))
                   (isTy _ τ⁻₁) (isTy _ τ⁻₂) =
          isTy _ (Π⁻ τ⁻₁ τ⁻₂)
  checkTy Γ (Elʳʷ e) with checkTm Γ (✶ Γ) e =>
    checkTy₃ _ _                    noTm        = noTy
    checkTy₃ Γ (Elʳʷ (eraseTm⁻ t⁻)) (isTm _ t⁻) = isTy _ (El⁻ t⁻)

  checkTm : (Γ : Ctxt) -> (τ : Ty Γ) -> (e : Tmʳʷ) -> τ ⊢Tm↑? e
  checkTm Γ τ e with inferTm Γ e =>
    checkTm₁ _ _ _             noTyTm          = noTm
    checkTm₁ Γ τ (eraseTm⁻ t⁻) (isTyTm σ _ t⁻) with σ ≟ τ =>
      checkTm₂ _ _ _             _               notEqualTy    = noTm
      checkTm₂ Γ τ (eraseTm⁻ t⁻) (isTyTm σ _ t⁻) (equalTy eqτ) =
        isTm _ (t⁻ ∷≛≡⊢⁻ eqτ)

  inferTm : (Γ : Ctxt) -> (e : Tmʳʷ) -> Γ ⊢Tm↓? e
  inferTm Γ (varʳʷ v) with inferVar Γ v =>
    inferTm₁ _ _                    noVar     = noTyTm
    inferTm₁ Γ (varʳʷ (eraseVar v)) (isVar v) = isTyTm _ _ (var⁻ Var v)
  inferTm Γ (λʳʷ e₁ e₂) with checkTy Γ e₁ =>
    inferTm₂ _ _                       noTy          = noTyTm
    inferTm₂ Γ (λʳʷ (eraseTy⁻ τ⁻₁) e₂) (isTy τ₁ τ⁻₁)
      with inferTm (Γ ▻ τ₁) e₂ =>
        inferTm₃ _ _ _ noTyTm = noTyTm
        inferTm₃ Γ (λʳʷ (eraseTy⁻ τ⁻₁) (eraseTm⁻ t⁻))
                   (isTy τ₁ τ⁻₁) (isTyTm τ₂ _ t⁻) =
          isTyTm (Π τ₁ τ₂) _ (λ⁻ τ⁻₁ t⁻)
  inferTm Γ (e₁ @ʳʷ e₂) with inferTm Γ e₁ =>
    inferTm₄ _ _                     noTyTm                   = noTyTm
    inferTm₄ Γ (eraseTm⁻ t⁻₁ @ʳʷ e₂) (isTyTm (Π τ₁ τ₂) _ t⁻₁)
      with checkTm Γ τ₁ e₂ =>
        inferTm₅ Γ (eraseTm⁻ t⁻₁ @ʳʷ eraseTm⁻ t⁻₂)
                   (isTyTm _ _ t⁻₁) (isTm _ t⁻₂) =
          isTyTm _ _ (t⁻₁ @⁻ t⁻₂)
        inferTm₅ _ _ _ noTm = noTyTm

------------------------------------------------------------------------
-- Completeness proof

-- Soundness follows from the types of the type checking functions.

mutual

  (=Var↓) (Γ : Ctxt) (τ : Ty Γ) (e : Varʳʷ)
    : (v : Γ ∋ τ) -> Γ ⊢Var↓? e -> Set
  (=Var↓) = EqVar↓ _ _ _

  data EqVar↓ :  (Γ : Ctxt) -> (τ : Ty Γ) -> (e : Varʳʷ)
              -> (v : Γ ∋ τ) -> Γ ⊢Var↓? e -> Set where
    isVar! (Γ : Ctxt) (τ¹, τ² : Ty Γ) (v¹ : Γ ∋ τ¹) (v² : Γ ∋ τ²)
           (isv? : Γ ⊢Var↓? eraseVar v²)
      : v¹ ≈∋ v² -> isv? ≡ isVar v² -> v¹ =Var↓ isv?

mutual

  (=Ty↑) (Γ : Ctxt) (τ : Ty Γ) (e : Tyʳʷ)
    : (τ⁻ : Ty⁻ τ) -> Γ ⊢Ty↑? e -> Set
  (=Ty↑) = EqTy↑ _ _ _

  data EqTy↑ :  (Γ : Ctxt) -> (τ : Ty Γ) -> (e : Tyʳʷ)
             -> (τ⁻ : Ty⁻ τ) -> Γ ⊢Ty↑? e -> Set where
    isTy! (Γ : Ctxt) (τ¹, τ² : Ty Γ) (τ⁻¹ : Ty⁻ τ¹) (τ⁻² : Ty⁻ τ²)
          (isτ⁻? : Γ ⊢Ty↑? eraseTy⁻ τ⁻²)
      : τ⁻¹ ≛⁻ τ⁻² -> isτ⁻? ≡ isTy τ² τ⁻² -> τ⁻¹ =Ty↑ isτ⁻?

mutual

  (=Tm↑) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (tv : TmVar) (e : Tmʳʷ)
    : (t⁻ : Tm⁻ tv t) -> τ ⊢Tm↑? e -> Set
  (=Tm↑) = EqTm↑ _ _ _ _ _

  data EqTm↑ :  (Γ : Ctxt) -> (τ : Ty Γ) -> (t : Γ ⊢ τ) -> (tv : TmVar)
             -> (e : Tmʳʷ)
             -> (t⁻ : Tm⁻ tv t) -> τ ⊢Tm↑? e -> Set where
    isTm! (Γ : Ctxt) (τ : Ty Γ)
          (t¹ : Γ ⊢ τ) (tv¹ : TmVar) (t⁻¹ : Tm⁻ tv¹ t¹)
          (t² : Γ ⊢ τ) (tv² : TmVar) (t⁻² : Tm⁻ tv² t²)
          (ist⁻? : τ ⊢Tm↑? eraseTm⁻ t⁻²)
      : t⁻¹ ≈⊢⁻ t⁻² -> ist⁻? ≡ isTm t² t⁻² -> t⁻¹ =Tm↑ ist⁻?

mutual

  (=Tm↓) (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (tv : TmVar) (e : Tmʳʷ)
    : (t⁻ : Tm⁻ tv t) -> Γ ⊢Tm↓? e -> Set
  (=Tm↓) = EqTm↓ _ _ _ _ _

  data EqTm↓ :  (Γ : Ctxt) -> (τ : Ty Γ) -> (t : Γ ⊢ τ) -> (tv : TmVar)
             -> (e : Tmʳʷ)
             -> (t⁻ : Tm⁻ tv t) -> Γ ⊢Tm↓? e -> Set where
    isTyTm! (Γ : Ctxt)
            (τ¹ : Ty Γ) (t¹ : Γ ⊢ τ¹) (tv¹ : TmVar) (t⁻¹ : Tm⁻ tv¹ t¹)
            (τ² : Ty Γ) (t² : Γ ⊢ τ²) (tv² : TmVar) (t⁻² : Tm⁻ tv² t²)
            (ist⁻? : Γ ⊢Tm↓? eraseTm⁻ t⁻²)
      : t⁻¹ ≈⊢⁻ t⁻² -> ist⁻? ≡ isTyTm τ² t² t⁻² -> t⁻¹ =Tm↓ ist⁻?

abstract

  inferVarComplete (Γ : Ctxt) (τ : Ty Γ)
    : (v : Γ ∋ τ) -> v =Var↓ inferVar Γ (eraseVar v)
  inferVarComplete (vz _)   = isVar! (reflVar $ vz _) refl
  inferVarComplete (vs v σ) with inferVarComplete v =>
    _ (vs v σ) (isVar! eqv eqv?) =
      rewrite-plugin(eqv?, isVar! (vsCong eqv (reflTy _)) refl)
  inferVarComplete (v ∷≡∋ eqτ) with inferVarComplete v =>
    _ (v ∷≡∋ eqτ) (isVar! eqv eqv?) =
      rewrite-plugin(eqv?, isVar! (castVarEqˡ eqτ eqv) refl)

mutual

  checkTyComplete (Γ : Ctxt) (τ : Ty Γ)
    : (τ⁻ : Ty⁻ τ) -> τ⁻ =Ty↑ checkTy Γ (eraseTy⁻ τ⁻)
  checkTyComplete (✶⁻ Γ)       = isTy! (reflTy⁻ $ ✶⁻ Γ) refl
  checkTyComplete (Π⁻ τ⁻₁ τ⁻₂) with checkTyComplete τ⁻₁ =>
    _ (Π⁻ τ⁻₁ τ⁻₂) (isTy! eqτ⁻₁ eqτ⁻₁?) with checkTyComplete τ⁻₂ =>
      _ (Π⁻ τ⁻₁ τ⁻₂) (isTy! eqτ⁻₁ eqτ⁻₁?) (isTy! eqτ⁻₂ eqτ⁻₂?) =
        rewrite-plugin(eqτ⁻₁?, rewrite-plugin(eqτ⁻₂?,
          isTy! (Π⁻Cong eqτ⁻₁ eqτ⁻₂) refl))
  checkTyComplete (El⁻ t⁻) with checkTmComplete t⁻ =>
    _ (El⁻ t⁻) (isTm! eqt⁻ eqt⁻?) =
      rewrite-plugin(eqt⁻?, isTy! (El⁻Cong eqt⁻) refl)
  checkTyComplete (τ⁻ ∷≡✶⁻ eqτ) with checkTyComplete τ⁻ =>
    _ (τ⁻ ∷≡✶⁻ eqτ) (isTy! eqτ⁻ eqτ⁻?) =
      rewrite-plugin(eqτ⁻?, isTy! (castTy⁻Eqˡ eqτ eqτ⁻) refl)

  checkTmComplete (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (tv : TmVar)
    : (t⁻ : Tm⁻ tv t) -> t⁻ =Tm↑ checkTm Γ τ (eraseTm⁻ t⁻)
  checkTmComplete t⁻ with inferTmComplete t⁻ =>
    _ t⁻ (isTyTm! eqt⁻ eqt⁻?) with tyEqComplete (tm⁻EqTyEq eqt⁻) =>
      _ t⁻ (isTyTm! eqt⁻ eqt⁻?) (tyEq! eqτ eqτ?) =
        rewrite-plugin(eqt⁻?, rewrite-plugin(eqτ?,
          isTm! (symTm⁻ $ castTm⁻Eq _ _) refl))

  inferTmComplete (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ) (tv : TmVar)
    : (t⁻ : Tm⁻ tv t) -> t⁻ =Tm↓ inferTm Γ (eraseTm⁻ t⁻)
  inferTmComplete (var⁻ _ v) with inferVarComplete v =>
    _ (var⁻ _ v) (isVar! eqv eqv?) =
      rewrite-plugin(eqv?, isTyTm! (var⁻Cong eqv) refl)
  inferTmComplete (λ⁻ τ⁻₁ t⁻) with checkTyComplete τ⁻₁ =>
    _ (λ⁻ τ⁻₁ t⁻) (isTy! eqτ⁻₁ eqτ⁻₁?) with inferTmComplete t⁻ =>
      _ (λ⁻ τ⁻₁ t⁻) (isTy! eqτ⁻₁ eqτ⁻₁?) (isTyTm! eqt⁻ eqt⁻?) =
        rewrite-plugin(eqτ⁻₁?, rewrite-plugin(eqt⁻?,
          isTyTm! (λ⁻Cong eqt⁻) refl))
  inferTmComplete (t⁻₁ @⁻ t⁻₂) with inferTmComplete t⁻₁ =>
    _ (t⁻₁ @⁻ t⁻₂) (isTyTm! eqt⁻₁ eqt⁻₁?) with checkTmComplete t⁻₂ =>
      _ (t⁻₁ @⁻ t⁻₂) (isTyTm! eqt⁻₁ eqt⁻₁?) (isTm! eqt⁻₂ eqt⁻₂?) =
        rewrite-plugin(eqt⁻₁?, rewrite-plugin(eqt⁻₂?,
          isTyTm! (app⁻Cong eqt⁻₁ eqt⁻₂) refl))
  inferTmComplete (t⁻ ∷≡⊢⁻ eqt) with inferTmComplete t⁻ =>
    _ (t⁻ ∷≡⊢⁻ eqt) (isTyTm! eqt⁻ eqt⁻?) =
      rewrite-plugin(eqt⁻?, isTyTm! (castTm⁻Eq t⁻ eqt) refl)

TypeChecker : Type
TypeChecker = time-plugin(TypeChecker, Set)
